#define MAX_WQ_ENTRIES 4
#define MAX_THREADS    3

chan msg = [0] of { byte, byte, byte };
chan rep[MAX_THREADS] = [1] of { byte };

typedef sem_wq_t
{
  short thread;
  byte  prio;
  short next;  // pointer replace by index into array
}

typedef cli_sem_t
{
  short pending;
  short queue;
}

short sem_counter;

sem_wq_t wq_entries[MAX_WQ_ENTRIES]
byte next_entry = 0;
cli_sem_t sem;  /*  maybe move init. to init*/

short enqueue_count = 0;
bool  tamer_is_waiting = false;
byte  temp;
byte  threads_outside = MAX_THREADS;
byte  threads_inside  = 0;

inline init_globals()
{
  d_step
  {
    sem_counter = 1;
    sem.pending = 0;
    sem.queue   = -1;

    temp = 0;
    do
      ::
         wq_entries[temp].thread = -1;
         temp = temp + 1;
         if
           :: (temp >= MAX_WQ_ENTRIES)
              -> break
           :: else;
         fi
    od
  }
}


inline ipc_wait(src, op, prio)
{
  tamer_is_waiting = true;
  msg?src,op,prio;
  tamer_is_waiting = false
}

inline ipc_reply_and_wait(src, op, prio)
{
  rep[src]!0;
  tamer_is_waiting = true;
  msg?src,op,prio;
  tamer_is_waiting = false
}

inline ipc_call(op, prio)
{
  msg!thread,op,prio;  /* send my thread id and my operation */
  rep[thread]?0        /* wait for reply */
}

inline wakeup_thread_without_switchto(src)
{
  rep[src]!0;
}

inline alloc_entry(entry)
{
  i = next_entry;
  do
    ::
       if
         :: (wq_entries[i].thread == -1)
            -> next_entry = (i + 1) % MAX_WQ_ENTRIES;
            entry = i;
            break;
         :: else
       fi;
       i = (i + 1) % MAX_WQ_ENTRIES;
       if
         :: (i == next_entry);
            -> assert(false)
       fi
  od
  /* just for statistics */
  ;enqueue_count++
  /* just for statistics */
}

inline free_entry(entry)
{
  d_step
  {
    wq_entries[entry].thread = -1
    /* just for statistics */
    ;enqueue_count--
    /* just for statistics */
  }
}

inline enqueue_thread(t, prio)
{
  d_step
  {
    alloc_entry(wq);

    wq_entries[wq].thread = t;
    wq_entries[wq].prio   = prio;
    wq_entries[wq].next   = -1;

    if
      :: (sem.queue == -1) -> sem.queue = wq;
      :: else -> wq_entries[wq].next = sem.queue;
                 sem.queue = wq;
    fi
  }
}

inline prio_highest(wq)
{
  d_step
  {
    wq_prio_highest = sem.queue;
    wp = wq_entries[wq_prio_highest].next;

    do
      ::
         if
           :: (wp != -1)
              -> if
                   :: (wq_entries[wp].prio > wq_entries[wq_prio_highest].prio)
                      -> wq_prio_highest = wp
                   :: else;
              fi;
              wp = wq_entries[wp].next
           :: else break;
         fi
    od;

    wq = wq_prio_highest;
  }
}


proctype Tamer()
{
  byte  operation;
  short wakeup;
  short src;
  byte  prio;
  short i;
  short wq;
  short wp;
  short wq_prio_highest;

  /* setup data structures */
no_reply:
/*progress2:*/
  ipc_wait(src, operation, prio);
  do
    ::
       if
         :: (operation == 1)  /* CLI */
            -> if
                 :: (sem.pending == 0)
                    -> enqueue_thread(src, prio);
                    goto no_reply;
                 :: else
                    -> sem.pending--;
                    sem_counter++;
            fi
         :: (operation == 2)  /* STI */
            -> if
              :: (sem.queue != -1)
                 -> enqueue_thread(src, prio);
                 prio_highest(wq_prio_highest);
                 do
                   :: (sem.queue != -1)
                      -> wp = sem.queue;
                      sem.queue = wq_entries[wp].next;
                      if
                        :: (wp != wq_prio_highest)
                           -> wakeup = wq_entries[wp].thread;
                           sem_counter++;
                           free_entry(wp);
                           wakeup_thread_without_switchto(wakeup);
                        :: else;
                      fi
                   :: else -> break;
                 od;
                 src = wq_entries[wq_prio_highest].thread;
                 free_entry(wq_prio_highest);
              :: else;
            fi;
            sem.pending = 1 - sem_counter;
         :: else -> assert(false);
       fi;
/*progress3:*/
       ipc_reply_and_wait(src, operation, prio);
  od
}

inline cli()
{
  do
    ::
       d_step
       {
         uf = 0;
         sem_counter--;
         if
           :: (sem_counter < 0) -> uf = 1;
           :: else;
         fi
       }

       if
         :: (uf == 1) -> ipc_call(1, prio);
         :: else -> break;
       fi
  od
}

inline sti()
{
  d_step
  {
    pf = 0;
    sem_counter++;
    if
      :: (sem_counter < 1) -> pf = 1;
      :: else;
    fi
  }

  if
    :: (pf == 1) -> ipc_call(2, prio);
    :: else;
  fi
}

proctype Thread(byte thread; byte prio)
{
  byte uf;
  byte pf;

  do
    ::
       threads_outside--;
       cli();
       threads_inside++;
progress1:
       threads_inside--;
       sti()
       threads_outside++;
  od
}

never
{
  do
    :: assert( ! ((enqueue_count + threads_outside >= MAX_THREADS) &&
                  (enqueue_count >= 1) &&
                  (tamer_is_waiting == true ) ||   /* Deadlock */
                  (threads_inside > 1)             /* Safety   */
    ))
  od
}

init
{
  init_globals();
  run Tamer();
  temp = 0;
  do
    ::
       run Thread(temp, temp);
       temp++;
       if
         :: (temp >= MAX_THREADS) -> break;
         :: else;
       fi
  od
}
